-- Solar2D backend for MoonNuklear - glbackend.lua
-- Derived from the original examples that come with Nuklear.

local nk = require("plugin.moonnuklear")
local Bytemap = require("plugin.Bytemap")
local serialize = require("plugin.serialize")

-- shortcuts
local floor = math.floor
local max = math.max
local min = math.min
local table_insert = table.insert

local s_unpack = serialize.struct.unpack

-------------------------------------------------------------------------------
local window, width, height, display_width, display_height, scale_x, scale_y
local ctx, atlas, cmds
local scroll_x, scroll_y, char_text
local last_button_click, double_click_pos, is_double_click_down
local vbo_size, ebo_size, anti_aliasing
local vbuf, ebuf
local mouse_grabbing, double_click_lo, double_click_hi
local circle_segment_count, curve_segment_count, arc_segment_count, global_alpha

-------------------------------------------------------------------------------
local function configure(ctx)
   cmds = nk.new_buffer("dynamic")

   -- struct vertex { float position[2]; float uv[2]; nk_byte col[4]; }
   local stride = 4*4 + 4*1
   local pos_offset, uv_offset, col_offset = 0, 2*4, 4*4

   -- fill convert configuration
   ctx:config_vertex_layout({
      { 'position', 'float', pos_offset },
      { 'texcoord', 'float', uv_offset },
      { 'color', 'r8g8b8a8', col_offset },
   })
   ctx:config_vertex_size(stride)
   ctx:config_vertex_alignment(4) --@@?
   ctx:config_circle_segment_count(circle_segment_count)
   ctx:config_curve_segment_count(curve_segment_count)
   ctx:config_arc_segment_count(arc_segment_count)
   ctx:config_global_alpha(global_alpha)
   if anti_aliasing then
      ctx:config_line_aa(true)
      ctx:config_shape_aa(true)
   end



   vbuf = nk.new_buffer("fixed")
   ebuf = nk.new_buffer("fixed")
end

-------------------------------------------------------------------------------
local mtex = graphics.newTexture{ type = "maskCanvas", width = 8, height = 8 }--, pixelWidth = 8, pixelHeight = 8 }
local mask_back = display.newRect(0, 0, 2, 2)
 
mtex:draw(mask_back)
mtex:invalidate()

local mask = graphics.newMask(mtex.filename, mtex.baseDir)

-------------------------------------------------------------------------------

local Images = {}

local Frame

local Colors = {}
local UVs = {}
local Verts = {}

local function render()
  display.remove(Frame)

  Frame = display.newGroup()

  local vertices = vbuf:init(--[[vertices, ]]vbo_size)
  local elements = ebuf:init(--[[elements, ]]ebo_size)

   ctx:convert(cmds, vbuf, ebuf)

   -- iterate over and execute each draw command
   local commands = ctx:draw_commands(cmds)
   local index, ipos, vpos, voff, coff = 0, 1, 1, 0, 0
   for i, cmd in ipairs(commands) do
      local elem_count = cmd.elem_count
      if elem_count > 0 then
         local x, y, w, h = unpack(cmd.clip_rect)

         x = floor(x * scale_x)
         y = floor((height - (y + h))*scale_y)
         w = floor(w * scale_x)
         h = floor(h * scale_y)

        local indices, min_index, max_index = {}, 1 / 0

        for _ = 1, elem_count / 3 do
          local i1, i2, i3, pos = s_unpack("HHH", elements, ipos)

          if i1 ~= 0 or i2 ~= 0 or i3 ~= 0 then
            indices[#indices + 1] = i1
            indices[#indices + 1] = i2
            indices[#indices + 1] = i3

            min_index = min(min_index, i1, i2, i3)
            max_index = max(max_index or -1, i1, i2, i3)
          end

          ipos = pos
        end
        
        if max_index then
          for _ = index, max_index do
            local x, y, u, v, r, g, b, a, pos = s_unpack("ffffBBBB", vertices, vpos)

            UVs[voff + 1], Verts[voff + 1] = u, x
            UVs[voff + 2], Verts[voff + 2] = v, y
            voff, vpos = voff + 2, pos

            Colors[coff + 1], Colors[coff + 2], Colors[coff + 3], Colors[coff + 4] = r / 255, g / 255, b / 255, a / 255
            coff = coff + 4
          end

          index = max_index

          local mesh = display.newMesh{
            parent = Frame,
            uvs = UVs, vertices = Verts, indices = indices,
            mode = "indexed", zeroBasedIndices = true
          }

          local cpos = min_index * 4
          
          for i = min_index, max_index do
            mesh:setFillVertexColor(i + 1, Colors[cpos + 1], Colors[cpos + 2], Colors[cpos + 3], Colors[cpos + 4])

            cpos = cpos + 4
          end
          
          mesh:translate(mesh.path:getVertexOffset())
          
          mesh.fill = Images[cmd.texture_id].fill
          --[[
          if x >= 0 and y >= 0 then
            mesh:setMask(mask)
print("??",x + w / 2,y + h / 2,mesh.x,mesh.y,mesh.path:getVertexOffset())
            mesh.maskX, mesh.maskScaleX = x + w / 2, w / 2
            mesh.maskY, mesh.maskScaleY = y + h / 2, h / 2
          end
          --]]
        end
      end
   end

   nk.clear_buffers(cmds, vbuf, ebuf) -- needed since Nuklear 4.00.0
   ctx:clear(ctx)
end

-------------------------------------------------------------------------------

local X, Y = display.contentCenterX, display.contentCenterY

local PrimaryDown, MiddleDown, SecondaryDown

Runtime:addEventListener("mouse", function(event)
  if event.type == "down" or event.type == "up" then
    PrimaryDown = event.isPrimaryButtonDown
    MiddleDown = event.isMiddleButtonDown
    SecondaryDown = event.isSecondaryButtonDown
  elseif event.type == "move" or event.type == "drag" then
    X, Y = event.x, event.y
  elseif event.type == "scroll" then
    scroll_x = scroll_x + event.scrollX
    scroll_y = scroll_y - event.scrollY
  end
end)

-------------------------------------------------------------------------------

local function char_callback(_, codepoint)
   table_insert(char_text, codepoint)
end

local function mouse_button_callback(window, button, action, shift, control, alt, super)
   if button ~= 'left' then return end
   if action == 'press' then 
      local now = system.getTimer()--glfw.get_time()
      local dt = now - last_button_click
      last_button_click = now
      if dt > double_click_lo and dt < double_click_hi then
         is_double_click_down = true
         double_click_pos = { X, Y }--glfw.get_cursor_pos(window) }
      end
   else 
      is_double_click_down = false
   end
end

local keys = {}
local keys_cb = {
   ['delete'] = function(down, control) keys.del = down end,
   ['enter'] = function(down, control) keys.enter = down end,
   ['tab'] = function(down, control) keys.tab = down end,
   ['backspace'] = function(down, control) keys.backspace = down end,
   ['up'] = function(down, control) keys.up = down end,
   ['down'] = function(down, control) keys.down = down end,
   ['home'] = function(down, control) keys.text_start, keys.scroll_start = down, down end,
   ['end'] = function(down, control) keys.text_end, keys.scroll_end = down, down end,
   ['page down'] = function(down, control) keys.scroll_down = down end,
   ['page up'] = function(down, control) keys.scroll_up = down end,
   ['left shift'] = function(down, control) keys.shift = down end,
   ['right shift'] = function(down, control) keys.shift = down end,
   ['c'] = function(down, control) keys.copy = down and control end,
   ['v'] = function(down, control) keys.paste = down and control end,
   ['x'] = function(down, control) keys.cut = down and control end,
   ['z'] = function(down, control) keys.text_undo = down and control end,
   ['r'] = function(down, control) keys.text_redo = down and control end,
   ['left'] = function(down, control) 
               if control then keys.text_word_left = down else keys.left = down end
            end,
   ['right'] = function(down, control)
               if control then keys.text_word_right = down else keys.right = down end
            end,
   ['b'] = function(down, control) keys.text_line_start = down and control end,
   ['e'] = function(down, control) keys.text_line_end = down and control end,
}

local function key_callback(window, key, scancode, action, shift, control, alt, super)
   local cb = keys_cb[key]
   if cb then cb(action ~= 'release', control) end
end

local function clipboard_paste(edit)
   local text = nil--glfw.get_clipboard_string(window)
   if #text>0 then edit:paste(text) end
end

local function clipboard_copy(text)
--   if #text > 0 then glfw.set_clipboard_string(window, text) end
end

-------------------------------------------------------------------------------
local function init(glfw_window, parameters)
-- Inits the backend with the given GLFW window (created with MoonGLFW) and
-- optional configuration parameters:
-- vbo_size          size of the vertex buffer object
-- ebo_size          size of the element buffer object
-- anti_aliasing     use anti-aliasing (default = false)
-- clipboard         use clipboard (default = false)
-- callbacks         install input callbacks (default = false)
-- mouse_grabbing    enable mouse grabbing (default = true)
-- double_click_lo   double click low threshold (default = .02 s)
-- double_click_hi   double click high threshold (default = .2 s)
-- circle_segment_count    no. of segments in circles (default = 22)
-- curve_segment_count     no. of segments in curves (default = 22)
-- arc_segment_count       no. of segments in arcs (default = 22)
-- global_alpha            global alpha (default = 1.0)
   assert(ctx == nil, "double init")
   local parameters = parameters or {}
   window = glfw_window
   vbo_size = parameters.vbo_size or 512*1024
   ebo_size = parameters.ebo_size or 128*1024
   anti_aliasing = parameters.anti_aliasing and true or false
   mouse_grabbing = parameters.mouse_grabbing==nil and true or parameters.mouse_grabbing
   double_click_lo = parameters.double_click_lo or 0.02
   double_click_hi = parameters.double_click_hi or 0.2
   circle_segment_count = parameters.circle_segment_count or 22
   curve_segment_count = parameters.curve_segment_count or 22
   arc_segment_count = parameters.arc_segment_count or 22
   global_alpha = parameters.global_alpha or 1.0

   last_button_click = 0
   is_double_click_down = false
   double_click_pos = {0, 0}
   char_text = {}
   scroll_x, scroll_y = 0, 0

   ctx = nk.init()

   if parameters.clipboard then
      ctx:set_clipboard_copy(clipboard_copy)
      ctx:set_clipboard_paste(clipboard_paste)
   end

   if parameters.callbacks then
   --   glfw.set_char_callback(window, char_callback)
   --   glfw.set_mouse_button_callback(window, mouse_button_callback)
   --   glfw.set_key_callback(window, key_callback)
   end

   configure(ctx)
   return ctx
end

-------------------------------------------------------------------------------
local function font_stash_begin()
    atlas = nk.new_font_atlas()
    atlas:begin()
    return atlas
end

local function font_stash_end(ctx, font)
   local image, w, h = atlas:bake('rgba32')
   local bm = Bytemap.newTexture{ width = w, height = h }

   bm:SetBytes(image)
   bm:PremultiplyAlpha()
   
   Images[#Images + 1] = { bm = bm, fill = { type = "image", filename = bm.filename, baseDir = bm.baseDir } }

   local null_tex, null_uv = atlas:done(#Images)
   ctx:config_null_texture(null_tex, null_uv)
   local font = font or atlas:default_font()
   if font then nk.style_set_font(ctx, font) end
end

-------------------------------------------------------------------------------

local function new_frame()

   width, height = display.contentWidth, display.contentHeight--[[ glfw.get_window_size(window)
   display_width, display_height = glfw.get_framebuffer_size(window)
]]
   scale_x = 1--display_width/width
   scale_y = 1--display_height/height

   nk.input_begin(ctx)
   if #char_text > 0 then nk.input_unicode(ctx, unpack(char_text)) end
   char_text = {}

   if mouse_grabbing then -- optional grabbing behavior
      if ctx:mouse_grab() then
      --   glfw.set_input_mode(window, 'cursor', 'hidden')
      elseif ctx:mouse_ungrab() then
      --   glfw.set_input_mode(window, 'cursor', 'normal')
      end
   end

   nk.input_keys(ctx, keys)

   local x, y = X, Y

   x = math.floor(x)
   y = math.floor(y)
   nk.input_motion(ctx, x, y)

   if mouse_grabbing and ctx:mouse_grabbed() then
      local prev_x, prev_y = ctx:mouse_prev()
    --  glfw.set_cursor_pos(window, prev_x, prev_y)
      nk.input_mouse_pos(ctx, prev_x, prev_y)
   end

   nk.input_button(ctx, 'left', x, y, PrimaryDown)
   nk.input_button(ctx, 'middle', x, y, MiddleDown)
   nk.input_button(ctx, 'right', x, y, SecondaryDown)

   nk.input_button(ctx, 'double', double_click_pos[1], double_click_pos[2], is_double_click_down)
   nk.input_scroll(ctx, scroll_x, scroll_y)
   scroll_x, scroll_y = 0, 0

   nk.input_end(ctx)
end

-------------------------------------------------------------------------------
local function shutdown()
   if ctx then
      atlas:free()
      ctx:free()
      cmds:free()
      -- Bytemaps, etc.
      ctx, atlas, cmds = nil
   end
end

-------------------------------------------------------------------------------
return {
   init = init,
   font_stash_begin = font_stash_begin,
   font_stash_end = font_stash_end,
   new_frame = new_frame,
   render = render,
   shutdown = shutdown,
   -- Callbacks:   
   scroll_callback = scroll_callback,
   char_callback = char_callback,
   mouse_button_callback = mouse_button_callback,
   key_callback = key_callback,
}

